/* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: alloc.c,v 1.5 2002/08/12 12:42:18 malekith Exp $
 *
 * Runtime environment for gont.
 *
 * Assumes compiler that allows $ in identifiers, such as GCC.
 * (It is required to have Ksi, so this ain't so important)
 */

#include <gont/runtime.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

//#undef GONT_USE_GC

#ifdef GONT_USE_GC
#  include <gc/gc.h>
#  define DEFAULT_CHUNK_SIZE 0
#else
#  define GC_malloc malloc
#  define GC_malloc_atomic malloc
#  define DEFAULT_CHUNK_SIZE (32*1024)
#endif

static void oops(int s) __attribute__ ((__noreturn__));

static void oops(int s)
{
	gont_oops("Run out of memory, trying to allocate %d bytes.\n", s);
}

static gont_ptroff chunk_size = DEFAULT_CHUNK_SIZE;
static gont_ptroff chunk_limit = 32;
static char *atomic_chunk_start, *atomic_chunk_end;
static char *normal_chunk_start, *normal_chunk_end;

void *gont_xmalloc(gont_ptroff s)
{
	void *p;
	
	if (chunk_size && s < chunk_limit) {
		if (s == 0)
			s = sizeof(gont_ptroff);
		else
			s = (s + sizeof(gont_ptroff) - 1) & 
					~(sizeof(gont_ptroff) - 1);
		p = normal_chunk_start;
		normal_chunk_start += s;
		if (normal_chunk_start > normal_chunk_end) {
			p = GC_malloc(chunk_size);
			if (p == NULL)
				oops(chunk_size);
			normal_chunk_end = (char*)p + chunk_size;
			normal_chunk_start = (char*)p + s;
		}

		return p;
	}
	
	p = GC_malloc(s);

	if (p == NULL) 
		oops(s);
	else
		return p;
}

void *gont_xmalloc0(gont_ptroff s)
{
	return memset(gont_xmalloc(s), '\0', s);
}


void *gont_xmalloc_atomic(gont_ptroff s)
{
	void *p;

	if (chunk_size && s < chunk_limit) {
		if (s == 0)
			s = sizeof(gont_ptroff);
		else
			s = (s + sizeof(gont_ptroff) - 1) & 
					~(sizeof(gont_ptroff) - 1);
		p = atomic_chunk_start;
		atomic_chunk_start += s;
		if (atomic_chunk_start > atomic_chunk_end) {
			p = GC_malloc_atomic(chunk_size);
			if (p == NULL)
				oops(chunk_size);
			atomic_chunk_end = (char*)p + chunk_size;
			atomic_chunk_start = (char*)p + s;
		}

		return p;
	}
	
	p = GC_malloc_atomic(s);

	if (p == NULL) 
		oops(s);
	else
		return p;
}

void *gont_xmalloc0_atomic(gont_ptroff s)
{
	return memset(gont_xmalloc_atomic(s), '\0', s);
}

void gont_init_memory_allocator()
{
	if (getenv("GONT_ALLOCATOR_CHUNK_SIZE"))
		chunk_size = atoi(getenv("GONT_ALLOCATOR_CHUNK_SIZE"));
	if (chunk_limit > chunk_size / 16)
		chunk_limit = chunk_size / 16;
}
