/* 
 * Copyright (c) 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: init.c,v 1.3 2002/06/04 11:54:08 malekith Exp $
 *
 * Runtime environment for gont.
 */

#include <gont/runtime.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

struct exit_chain {
	struct exit_chain *next;
	void (*fnc)(void*);
	void *data;
};

static struct exit_chain *exit_chain;

int gont_argc;
char **gont_argv;

void gont_init_runtime(void)
{
	gont_init_memory_allocator();
}

void gont_exit(void)
{
	exit(0);
}

void gont_set_argv(int argc, char **argv)
{
	gont_argc = argc;
	gont_argv = argv;
}

void gont_oops(const char *msg, ...)
{
	va_list ap;

	fputs("Gont runtime fatal error: ", stderr);
	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	va_end(ap);
	fputs("\n", stderr);

	abort();
}

void gont_run_exit_routines(void)
{
	struct exit_chain *p;
	
	while (exit_chain) {
		p = exit_chain;
		exit_chain = exit_chain->next;
		p->fnc(p->data);
	}
}

void gont_register_exit_routine(void (*fnc)(void*), void *data)
{
	struct exit_chain *c;

	c = gont_xmalloc(sizeof(*c));
	c->next = exit_chain;
	c->fnc = fnc;
	c->data = data;
	exit_chain = c;
}
